#include <stdio.h>
#include "c45_kick_dir.h"


/*#define DEBUG 1*/







/*++++++++++++++++++++++++++++++++++++++++++++++++

  c45_is_visible()
  ****************

  "In-house" test for visibility in terms of
  the C4.5 data structures.

  INPUT: "fval" is the value of the angle
         "at which" the object is visibile
	 wrt the player that has the ball.

  OUTPUT: TRUE/FALSE

++++++++++++++++++++++++++++++++++++++++++++++++*/

int c45_is_visible( float fval )
{
  if( fval >=- 1000000.00 )
    return 0;

  return 1;
}







/*++++++++++++++++++++++++++++++++++++++++++++++++

  GetKickingAngle()
  *****************

  Computes the *best* KICKING DIRECTION. In 
  order to do it, if converts the "state" info
  into some easier to use data, and then it 
  calls "ComputeKickingAngle()".

  INPUT: -

  OUTPUT: best KICKING DIRECTION in order to score.
  (you wish you knew it, right?)

++++++++++++++++++++++++++++++++++++++++++++++++*/

float GetKickingAngle()
{
  /* 40 = 21 +12 + 4 + 1 + 1 + 1 "c4.5 attribs" */
  float Players[_MAX_VISIBLE_PLAYERS_];
  float Flags[_MAX_FLAGS_];
  float Lines[_MAX_LINES_];
  float Ball;

  int cPlayerCnt;

  int i, j;


  /*-----------------------------*/
  /*##### INIT DATA STRUCTS #####*/
  
  for( i=0 ; i<_MAX_VISIBLE_PLAYERS_ ; i++ )
    Players[i] = _INVISIBLE_;
  
  for( i=0 ; i<_MAX_FLAGS_ ; i++ )
    Flags[i] = _INVISIBLE_;
  
  for( i=0 ; i<_MAX_LINES_ ; i++ )
    Lines[i] = _INVISIBLE_;
  
  Ball = _INVISIBLE_;
  cPlayerCnt = 0;



  /*------------------------------------*/
  /*###### set all relevant values #####*/

  cPlayerCnt = State[NOW].Auxilary.cPlayerCnt;

  for( i=0 ; i<cPlayerCnt ; i++ )
    /* DOESN'T need any test because ALL visible players are
     *consecutive* in "AnglesOfVisiblePlayers[]" */
    Players[i] = State[NOW].Auxilary.AnglesOfVisiblePlayers[i];

  for( i=0 ; i<_MAX_FLAGS_ ; i++ )
    if( State[NOW].Flags[i].visible != FALSE )
      Flags[i] = State[NOW].Flags[i].SeenInfo.dir;
  
  for( i=0 ; i<_MAX_LINES_ ; i++ )
    if( State[NOW].Lines[i].visible != FALSE )
      Lines[i] = State[NOW].Lines[i].SeenInfo.dir;
  
  if( State[NOW].Ball.visible != FALSE )
    /* BAD !!! */
    Ball = State[NOW].Ball.SeenInfo.dir;

      
#ifdef DEBUG			  
  printf("old GD = %f\t new GD = %f\n", 
	 State[NOW].Goals[OPPONENT].SeenInfo.dist,
	 State[NOW].Goals[OPPONENT].SeenInfo.dist *
	 cos( (State[NOW].Goals[OPPONENT].SeenInfo.dir+
	       State[NOW].Players[OWN][MyNumber-1].position.dir)	       
	      *PI/180.0));
#endif

  return ComputeKickingAngle( Players, cPlayerCnt, 
			      Flags, Lines, Ball,
			      State[NOW].Players[OWN][MyNumber-1].position.dir, 
			      State[NOW].Goals[OPPONENT].SeenInfo.dist * 
			      cos( (State[NOW].Goals[OPPONENT].SeenInfo.dir+
				    State[NOW].Players[OWN][MyNumber-1].position.dir)
				   *PI/180.0), 
			      State[NOW].Goals[OPPONENT].SeenInfo.dir);
}








/*++++++++++++++++++++++++++++++++++++++++++++++++

  KickingAngle()
  *****************

  Creates a C4.5 case, obtains a classification
  for it (i.e., a "general direction" UP/DOWN/CENTER), 
  and then it calls "LowLevelComputeAngle()" to get 
  the angle that corresponds to the "general direction".

  INPUT: bunch of floats (and arrays of floats) 
  that represent the angles of the visible PLAYERS
  (+ an INT for NUMBER OF VISIBLE PLAYERS), FLAGS,
  LINES, BALL, the PLAYER-WITH-THE-BALL_FACING_DIRECTION,
  and both the goal's DIRECTION and DISTANCE.

  OUTPUT: best KICKING DIRECTION in order to score.
  (you wish you knew it, right?)

++++++++++++++++++++++++++++++++++++++++++++++++*/

float ComputeKickingAngle( float* Players, int cPlayerCnt, float* Flags, 
			   float* Lines, float Ball, float facing_dir, 
			   float goal_dist, float goal_dir )
{
  char* pszClasiffierAnswer;
  float final_angle;
  char pszNewCase[2000];
  int i, j;


  /*-----------------------*/
  /*##### CREATE CASE #####*/
  for( j=0 ; j<21 ; j++ )
    {
      float tmp;
	  
      for( i=0 ; i<21 ; i++ )
	if( ! c45_is_visible(Players[i]) )
	  sprintf(pszNewCase, "%f, ", Players[i]);
	else
	  sprintf(pszNewCase, "?, ");
      
      for( i=0 ; i<12 ; i++ )
	if( ! c45_is_visible(Flags[i]) )
	  sprintf(pszNewCase, "%f, ", Flags[i]);
	else
	  sprintf(pszNewCase, "?, ");
            
      for( i=0 ; i<4 ; i++ )
	if( ! c45_is_visible(Lines[i]) )
	  sprintf(pszNewCase, "%f, ", Lines[i]);
	else
	  sprintf(pszNewCase, "?, ");
      
      if( ! c45_is_visible(Ball) )
	sprintf( pszNewCase, "%f, ", Ball);
      else
	sprintf(pszNewCase, "?, ");
    }


  pszClasiffierAnswer = classify(pszNewCase, _KICK_TO_SCORE_);
  reset();
  
  if( !strcmp( pszClasiffierAnswer, _UP_) )
    final_angle = LowLevelComputeAngle( goal_dist, goal_dir, 
				       facing_dir, _KICK_UP_);
  else
    if( !strcmp( pszClasiffierAnswer, _DOWN_) )
      final_angle = LowLevelComputeAngle( goal_dist, goal_dir, 
					 facing_dir, _KICK_DOWN_);
    else
      final_angle = LowLevelComputeAngle(goal_dist, goal_dir, 
					facing_dir, _KICK_CENTER_);
}






/*++++++++++++++++++++++++++++++++++++++++++++++++

  LowLevelComputeAngle()
  **********************

  Creates a C4.5 case, obtains a classification
  for it (i.e., a "general direction" UP/DOWN/CENTER), 
  and then it calls "LowLevelComputeAngle()" to get 
  the angle that corresponds to the "general direction".

  INPUT: 
  ++++++

  goal_dist: distance from PLAYER-WITH-BALL to Opponent's goal.
  rel_goal_dir: angle of Opponent's goal wrt PLAYER-WITH-BALL.
  facing_dir: facing-angle to the PLAYER-WITH-BALL wrt
              the ATTACKING DIRECTION
  raw_kick_dir: UP/DOWN/CENTER


  OUTPUT: best KICKING DIRECTION in order to score.
  +++++++ That is, AN ABSOLUTE ANGLE wrt the ATTACKING DIRECTION.

++++++++++++++++++++++++++++++++++++++++++++++++*/


float LowLevelComputeAngle( float goal_dist, float rel_goal_dir, 
			    float facing_dir, int raw_kick_dir)
{
  float goal_dir = rel_goal_dir + facing_dir; 

  double angle;

#ifdef DEBUG			  
  printf("\n +++ goal_dir = %f  rgd=%f  fd=%f\n", goal_dir, rel_goal_dir, facing_dir); 
#endif

  if( goal_dir != 0.0 )
    {
      switch( raw_kick_dir )
	{
	case _KICK_UP_:
	  if( goal_dir > 0 )
	    {
	      angle = (180.0 * atan( ((goal_dist*tan(goal_dir*PI/180.0)) -
				      ((14.02)*0.25)) / goal_dist)) / PI;
#ifdef DEBUG			  
	      printf("\n +++ d=%f x=%f, y=%f, z=%f", goal_dist,
		     goal_dist, (goal_dist*tan(goal_dir*PI/180.0)),
		     ((goal_dist*tan(goal_dir*PI/180.0)) - ((14.02)*0.25)));
#endif
	    }
	  else
	    {
	      angle = (-180.0 * 
		       atan( (fabs((goal_dist*tan(goal_dir*PI/180.0))) +
			      (0.25*(14.02)))
			     / goal_dist)) / PI;
	      
#ifdef DEBUG			  
	      printf("\n --- d=%f x=%f, y=%f, z=%f", goal_dist,
		     goal_dist,
		     (goal_dist*tan(goal_dir*PI/180.0)),
		     (goal_dist*tan(goal_dir*PI/180.0))  + ((14.02)*0.25));
#endif
	    }

	  angle -= facing_dir;

#ifdef DEBUG			  
	  printf("\n\n ***** ANGLE (---up---) = %f\n", angle);
#endif
	  break;
		      
		      
		      
	case _KICK_DOWN_:
	  if( goal_dir > 0 )
	    {
	      angle = (180.0 * 
		       atan( ((goal_dist*tan(goal_dir*PI/180.0)) +
			      ((14.02)/4.0)) / goal_dist)) / PI;
	      
#ifdef DEBUG			  
	      printf("\n +++ d=%f x=%f, y=%f, z=%f", goal_dist,
		     goal_dist, ((goal_dist*tan(goal_dir*PI/180.0))),
		     ((goal_dist*tan(goal_dir*PI/180.0)) + ((14.02)/4.0)));
#endif
	    }
	  else
	    {
	      angle = (-180.0 * 
		       atan( (fabs(goal_dist*tan(goal_dir*PI/180.0)) -
			      ((14.02)/4.0)) / goal_dist)) / PI;

#ifdef DEBUG			  
	      printf("\n --- edge-1 = %f\t, goal_dist=%f, angle-r=%f, angle-d=%f\n", 
		     (fabs(goal_dist*tan(goal_dir*PI/180.0)) - ((14.02)/4.0)),
		     goal_dist,
		     atan((fabs(goal_dist*tan(goal_dir*PI/180.0)) - ((14.02)/4.0))/goal_dist),
		     (-180.0 * 
		     atan((fabs(goal_dist*tan(goal_dir*PI/180.0)) - ((14.02)/4.0))/goal_dist))/PI);
#endif
	    }
		      
	  angle -= facing_dir;

#ifdef DEBUG			  
	  printf("\n\n ***** ANGLE (---down---) = %f\n", angle);
#endif
	  break;
		      
	case _KICK_CENTER_:
	  angle = goal_dir;

#ifdef DEBUG			  	
	  printf("\n\n ***** ANGLE (---center---) = %f\n", angle);
#endif
	  break;
	}
    }
  else
    {
      switch( raw_kick_dir )
	{
	case _KICK_UP_:
	  angle = (180.0 * atan((- (14.02)/4.0)/goal_dist))/PI;

#ifdef DEBUG			  
	  printf("\n\n ***** ANGLE-0 (---up---) = %f\n", angle);
	  printf("\n d=%f x=%f, y=%f, gw=%f, z=%f", goal_dist,
		 goal_dist, 0.0, PI, -((14.02)/4.0));
#endif
	  break;
		      
	case _KICK_DOWN_:
	  angle = (180.0 * atan(((14.02)/4.0)/goal_dist))/PI;

#ifdef DEBUG			  
	  printf("\n\n ***** ANGLE-0 (---down---) = %f\n", angle);
	  printf("\n d=%f x=%f, y=%f, gw=%f, z=%f", goal_dist,
		 goal_dist, 0.0, PI, ((14.02)/4.0));
#endif
	  break;
		      
	case _KICK_CENTER_:
	  angle = goal_dir;

#ifdef DEBUG			  
	  printf("\n\n ***** ANGLE-0 (---center---) = %f\n", angle);
#endif
	  break;
	}
    }

  return angle;
}

