/* File   : kick_dir.c
   Info   : finds the best direction to kick the ball
   Author : Ali Erdem
   Date   : June 27, 1997
*/

#include "geometry.h"

/* the weights of desires to kick the ball in a certain direction */

#define NORMAL_UNKNOWN    1
#define NORMAL_KNOWN      2
#define FRIEND    2
#define FOE      -50
#define OWN_GOAL        -1000
/* #define OWN_GOAL_BAND_1 -10 */
/* #define OWN_GOAL_BAND_2 (OWN_GOAL_BAND_1/100) */

#define OPP_GOAL  0
/* #define OPP_GOAL_BAND_1     3 */
/* #define OPP_GOAL_BAND_2  (OPP_GOAL_BAND_1/5) */

#define SIMPLE_FORBIDDEN -1

#define OWN_SAFE_ZONE  30  /* amount of space to leave around own goal alone */

#define WEIGHT   0  /* indexes to array below */
#define KNOWIT   1

#define ARRAY_SIZE  360

/* the array for partitioning the field, it's a list of desire to kick for from-to angles  */ 
static int regions[ARRAY_SIZE][2];  /* 360 degrees, KNOWIT/WEIGHT */

static void set_regions (double lo, double hi, int W)
{
  int i, low, high;

  if (lo > hi)
    {
      int tmp = lo;
      lo = hi;
      hi = tmp;
    }

  low  = min(lo+180.0, 359);
  high = min(hi+180.0, 359);

  /* check that object is not behind us */
  if (high-low>180) /* if difference is greater, object is behind us */
    {
      for (i=high; (i%ARRAY_SIZE)!=low; i++)
	{
	  regions[i%ARRAY_SIZE][WEIGHT] = W;
	  regions[i%ARRAY_SIZE][KNOWIT] = TRUE;
	}
    }
  else
    {
      for(i=low; i<= high; i++)
	{
	  regions[i][WEIGHT] = W;
	  regions[i][KNOWIT] = TRUE;
	}
    }

    return ;
}

static void update_regions (double lo, double hi, int W)
{
  int i, low, high;

  if (lo > hi)
    {
      int tmp = lo;
      lo = hi;
      hi = tmp;
    }

  low  = min(lo+180.0, 359);
  high = min(hi+180.0, 359);

  /* check that object is not behind us */
  if (high-low>180) /* if difference is greater, object is behind us */
    {
      for (i=high; (i%ARRAY_SIZE)!=low; i++)
	{
	  regions[i%ARRAY_SIZE][WEIGHT] += W;
	  regions[i%ARRAY_SIZE][KNOWIT] = TRUE;
	}
    }
  else
    {
      for(i=low; i<= high; i++)
	{
	  regions[i][WEIGHT] += W;
	  regions[i][KNOWIT] = TRUE;
	}
    }

    return ;
}

void regions_init(void)
{
  bzero(regions, sizeof(regions));
  set_regions(-90,+90,NORMAL_UNKNOWN);
  set_regions(-45,+45,NORMAL_KNOWN);


  return;
}

static double find_best_region (void)
{
    int i=0,j=0;
    int cog1 = 0;
    int cog2 = 0;
    int max = 0;
    double cog = 0.0;

#undef   COG  /* Center of Gravity */
#undef   LCOG /* COG of largest area */
#undef  MAX  /* max defuz */
#define SPECIAL_LCOG /* LCOG that ignores objects in back */

#ifdef SPECIAL_LCOG
    /* XXX cog of largest area in -90 to 90 Major future bug... ;-) */
    max = 0;
    for (i=90; i<180; i++)
      {
	if (  (regions[i][KNOWIT])
	    &&(regions[i][WEIGHT] > 0))
	  {
	    cog1 = 0;
	    cog2 = 0;
	    for (j=i;
		   (regions[j][KNOWIT])
		 &&(regions[j][WEIGHT]>0)
		 &&(j<ARRAY_SIZE);
		 j++)
	      {
		cog1 += regions[j][WEIGHT]*j;
		cog2 += regions[j][WEIGHT];
	      }

	    if (cog2>max)
	      {
		cog = (double)cog1/(double)cog2 - 180.0;
		max = cog2;
	      }
	    i=j-1;
	  }
      }
    /* check for special case where no positive area found */
    if (0 == max)
      {
	return opponent_goal_center();
      }

    return cog;
#endif /* SPECIAL_LCOG */

#ifdef MAX
    max = -32000;
    j = 0;
    for (i=0;i<ARRAY_SIZE; i++)
      {
	if (  (regions[i][KNOWIT])
	    &&(regions[i][WEIGHT] > max))
	  {
	    max = regions[i][WEIGHT];
	    j = i;
	  }
      }
	
    return (double)(j-180);
#endif
#ifdef LCOG
    /* check for special case where an object is behind us, and so its positive
       weight starts somewhere just below 359, and ends just above 0 - basically,
       the case where the object's area is split over the start and end of array */
    max = 0;
    if (  (regions[ARRAY_SIZE-1][KNOWIT])     /* if positive at end of array */
	&&(regions[ARRAY_SIZE-1][WEIGHT] > 0)
	&&(regions[0][KNOWIT])                /* and positive at beginning of array */
	&&(regions[0][WEIGHT] > 0))
      {
	/* the treat as one area */

	/* search for beginning of this area */
	for (i=ARRAY_SIZE-1; 
	     (regions[i][KNOWIT]) && (regions[i][WEIGHT] > 0);
	     i--)
	  ;

	cog1 = 0;
	cog2 = 0;
		
	for(j=i+1;  /* first positive of area found */
	    ((j%ARRAY_SIZE) != i)    /* we don't hit i again from around... */
	    && (regions[j%ARRAY_SIZE][KNOWIT])
	    && (regions[j%ARRAY_SIZE][WEIGHT] > 0);
	    j++)
	  {
	    cog1 += regions[j%ARRAY_SIZE][WEIGHT]*(j%ARRAY_SIZE);
	    cog2 += regions[j%ARRAY_SIZE][WEIGHT];
	  }

	if (cog2>max)
	  {
	    cog = (double)cog1/(double)cog2 - 180.0;
	    max = cog2;
	  }
	fprintf(stderr,"Found area in my back: %d %d %2.1lf\n",j%ARRAY_SIZE-1-180,i+1-180, cog);

      }

    /* NOTE: (GalK)
       if one looks carefully, he/she will notice that the next section will again
       re-examine and test those areas we may have just covered in the special case
       above.  However, since they will be dealt with seperately, i.e., the area
       at beginning of array will be dealt with seperately than area at end, they
       (if they exist) will always be smaller than their combined sum.  So we
       don't have to worry about skipping them, although of course it is less
       elegant and efficient this way.  But who's interested in elegance or
       effiency at 1am? ;-)  */

    /* cog of largest area */
    max = 0;
    for (i=0; i<ARRAY_SIZE; i++)
      {
	if (  (regions[i][KNOWIT])
	    &&(regions[i][WEIGHT] > 0))
	  {
	    cog1 = 0;
	    cog2 = 0;
	    for (j=i;
		   (regions[j][KNOWIT])
		 &&(regions[j][WEIGHT]>0)
		 &&(j<ARRAY_SIZE);
		 j++)
	      {
		cog1 += regions[j][WEIGHT]*j;
		cog2 += regions[j][WEIGHT];
	      }

	    if (cog2>max)
	      {
		cog = (double)cog1/(double)cog2 - 180.0;
		max = cog2;
	      }
	    i=j-1;
	  }
      }

    /* check for special case where no positive area found */
    if (0 == max)
      {
	fprintf(stderr,"clear_ball.c/find_best_direction()/LCOG: No positive area found.\n");
	return opponent_goal_center();
      }


    /*    fprintf(stderr,"========= COG yields: %lf2.1\n", cog); */
    return cog;
#endif /* LCOG */

#ifdef COG  
    for (i=0;i<ARRAY_SIZE; i++)
      {
	if (regions[i][KNOWIT])
	  {
	    cog1 += regions[i][WEIGHT]*i;
	    cog2 += regions[i][WEIGHT];
	  }
      }

      fprintf(stderr,"========= COG yields: %lf2.1\n", (double)(cog1/cog2)); 
    return cog1/cog2;
#endif COG
}

void update_opponent_player(double angle, double distance)
{
  double p_zone, lo, hi;
  /* Amount of space to leave around a player */
#define P_RADIUS 5
  /* p_zone = amount to leave around a player in angles */
  /* note that p_zone is actually atan(P_RADIUS/sqrt(sqr(distance)-sqr(P_RADIUS)) */
  p_zone = Rad2Deg(atan(P_RADIUS/distance));
#undef P_RADIUS  
  lo = NormalizeAngle(angle-p_zone);
  hi = NormalizeAngle(lo+2*p_zone);
  
  update_regions(lo,hi,FOE);

  return ;
}

void update_team_player(double angle, double distance)
{
  double p_zone, lo, hi;
  /* Amount of space to leave around a player */
#define P_RADIUS 2
  /* p_zone = amount to leave around a player in angles */
  /* note that p_zone is actually atan(P_RADIUS/sqrt(sqr(distance)-sqr(P_RADIUS)) */
  p_zone = Rad2Deg(atan(P_RADIUS/distance));
#undef P_RADIUS  
  lo = NormalizeAngle(angle-p_zone);
  hi = NormalizeAngle(lo+2*p_zone);
  
  update_regions(lo,hi,FRIEND);

  return ;
}


double clear_ball_angle(void)
{
   static double best;
   int i,j;
   double lo,hi,tmp,p_zone;

   /* check all players */
   /* updated when seen in parser.c */

   /* opponent goal is between (field_length/2,-goal_width/2) and (field_length/2,+goal_width/2)  
   the center of the goal is (field_length/2,0) */

   lo = NormalizeAngle(Rad2Deg(atan2(-GOAL_WIDTH/2-State[NOW].Players[OWN][MyNumber-1].position.pos.y,
                                 FIELD_LENGTH/2-State[NOW].Players[OWN][MyNumber-1].position.pos.x))
     - State[NOW].Players[OWN][MyNumber-1].position.dir);

   hi = NormalizeAngle(Rad2Deg(atan2(GOAL_WIDTH/2-State[NOW].Players[OWN][MyNumber-1].position.pos.y,
                                 FIELD_LENGTH/2-State[NOW].Players[OWN][MyNumber-1].position.pos.x))
     - State[NOW].Players[OWN][MyNumber-1].position.dir);

   /* maximum to goal itself */
   update_regions(lo,hi,OPP_GOAL);
#ifdef OPP_GOAL_BAND_1
   /* than fuzzy for the area */
   update_regions(NormalizeAngle(lo-15), NormalizeAngle(hi+15), OPP_GOAL_BAND_1);
#ifdef OPP_GOAL_BAND_2
   update_regions(NormalizeAngle(lo-50), NormalizeAngle(hi+50), OPP_GOAL_BAND_2);
#endif
#endif

   /* own goal is between (-field_length/2,-goal_width/2) 
      and (-field_length/2,+goal_width/2)  
      the center of the goal is (field_length/2,0) */
   
   hi = NormalizeAngle(Rad2Deg(atan2(-(GOAL_WIDTH/2+OWN_SAFE_ZONE)-State[NOW].Players[OWN][MyNumber-1].position.pos.y,
				     -FIELD_LENGTH/2-State[NOW].Players[OWN][MyNumber-1].position.pos.x))
		       - State[NOW].Players[OWN][MyNumber-1].position.dir);
   
   lo = NormalizeAngle(Rad2Deg(atan2((GOAL_WIDTH/2+OWN_SAFE_ZONE)-State[NOW].Players[OWN][MyNumber-1].position.pos.y,
				     -FIELD_LENGTH/2-State[NOW].Players[OWN][MyNumber-1].position.pos.x))
		       - State[NOW].Players[OWN][MyNumber-1].position.dir);

   /* maximum penalty for own goal */
   update_regions(lo,hi,OWN_GOAL);
#ifdef OWN_GOAL_BAND_1
   /* then fuzzy for the area */
   update_regions(NormalizeAngle(lo-15),NormalizeAngle(hi+15),OWN_GOAL_BAND_1);
#ifdef OWN_GOAL_BAND_2
   update_regions(NormalizeAngle(lo-45),NormalizeAngle(hi+45),OWN_GOAL_BAND_2);
#endif
#endif   

   /* adjust for what we can actually kick */
   if (visible(State[NOW].Ball))
     {
       hi = NormalizeAngle(State[NOW].Ball.SeenInfo.dir - PROBLEM_ANGLE);
       lo = NormalizeAngle(State[NOW].Ball.SeenInfo.dir + PROBLEM_ANGLE);
       
       set_regions(lo, hi, SIMPLE_FORBIDDEN);
     }

   best = find_best_region();

   /*   fprintf(stderr,"best direction to kick=%7.2lf\n",best); */
   return best;
}
