/*************************************************************************/
/*								   	 */
/*	Classify items interactively using a set of rules		 */
/*	-------------------------------------------------		 */
/*								   	 */
/*************************************************************************/

#include <stdio.h>
#include "defns.i"
#include "types.i"

void TestMemAlloc();

/*  External data  */

extern char* pszInString;
extern FILE* Rf;


short		MaxAtt, MaxClass, MaxDiscrVal;

ItemNo		MaxItem;
ItemNo          cCurrentItem;

Description	*Item;

DiscrValue	*MaxAttVal;

String		*ClassName,
		*AttName,
		**AttValName,
                FileName; 

/*		FileName = "DF"; */

char		*SpecialStatus;

short		VERBOSITY = 0,
		TRACE     = 0;


Boolean		FirstTime = true;

PR		*Rule;

RuleNo		NRules = 0;
short		RuleSpace = 0;



short RuleSpaceStart, RuleSpaceEnd;
RuleNo NRulesStart, NRulesEnd; 
PR * RuleStart;
PR * RuleEnd;


extern Test    *TestVec;
extern short   NTests;
extern short   TestSpace;

extern short   MaxClass;
extern int     AttCeiling, ClassCeiling;

ClassNo		DefaultClass;


	/*  The interview module uses a more complex description of an
	    case called a "Range Description" with lower and upper bounds.
	    For further information, see consult.c  */

typedef	struct ValRange *RangeDescRec;

struct ValRange
	{
	    Boolean Known, Asked;
	    float LowerBound, UpperBound, *Probability;
	};

RangeDescRec RangeDesc;

float Confidence;

#define	MINCF	0.50		/* minimum cf for useable rule */


/*************************************************************************/
/*								  	 */
/*  Find the best rule for the current case.			  	 */
/*  Note:  leave probability in Confidence				 */
/*								  	 */
/*************************************************************************/


RuleNo BestRule()
/*     -------- 	 */
{
    RuleNo r;
    float cf, RuleStrength();

    Confidence = 0.0;
    
    ForEach(r, 1, NRules)
    {
	cf = RuleStrength(Rule[r]);

	if ( cf > 0.5 )
	{
	    Confidence = cf;
	    return r;
	}
    }

    return 0;
}



/*************************************************************************/
/*								    	 */
/*  Given a rule, determine the strength with which we can conclude	 */
/*  that the current case belongs to the class specified by the RHS	 */
/*  of the rule								 */
/*								    	 */
/*************************************************************************/


float RuleStrength(Rule)
/*    ------------ 	 */
    PR Rule;
{
    short d;
    float RuleProb=1.0, ProbSatisfied();

    ForEach(d, 1, Rule.Size)
    {
      
	RuleProb *= ProbSatisfied(Rule.Lhs[d]);
	if ( RuleProb < MINCF )
	{
	    return 0.0;
	}
    }

    return ( (1 - Rule.Error) * RuleProb );
}



/*************************************************************************/
/*							   		 */
/*  Determine the probability of the current case description		 */
/*  satisfying the given condition					 */
/*							   		 */
/*************************************************************************/


float ProbSatisfied(c)
/*    ------------- 	 */
    Condition c;
{
    Attribute a;
    char v;
    float AddProb=0.0;
    Test t;
    DiscrValue i;

    t = c->CondTest;
    a = t->Tested;
    v = c->TestValue;

    CheckValue(a);

    if ( ! RangeDesc[a].Known )
    {
	return 0.0;
    }

    switch ( t->NodeType )
    {
	case BrDiscr:  /* test of discrete attribute */

	    return RangeDesc[a].Probability[v];

	case ThreshContin:  /* test of continuous attribute */

	    if ( RangeDesc[a].UpperBound <= t->Cut )
	    {
		return ( v == 1 ? 1.0 : 0.0 );
	    }
	    else
	    if ( RangeDesc[a].LowerBound > t->Cut )
	    {
		return ( v == 2 ? 1.0 : 0.0 );
	    }
	    else
	    if ( v == 1 )
	    {
		return (t->Cut - RangeDesc[a].LowerBound) /
		       (RangeDesc[a].UpperBound - RangeDesc[a].LowerBound);
	    }
	    else
	    {
		return (RangeDesc[a].UpperBound - t->Cut) /
		       (RangeDesc[a].UpperBound - RangeDesc[a].LowerBound);
	    }

	case BrSubset:  /* subset test on discrete attribute  */

	    ForEach(i, 1, MaxAttVal[a])
	    {
		if ( In(i, t->Subset[v]) )
		{
		    AddProb += RangeDesc[a].Probability[i];
		}
	    }
	    return AddProb;

    } 
    return 0.0;
}



/*************************************************************************/
/*								  	 */
/*		Process a single case				  	 */
/*								  	 */
/*************************************************************************/


char* InterpretRuleset()
{ 
  char Reply;
  Attribute a;
  RuleNo r;

  ForEach(a, 0, MaxAtt)
    RangeDesc[a].Asked = false;


  /*
   * if ( r = BestRule() )
   * printf("Decision:\t%s  (CF = %.2f)\n", 
   * ClassName[Rule[r].Rhs], Confidence);
   * else
   * printf("Decision:\t%s (default class)\n", ClassName[DefaultClass]);
   */
    
  if ( r = BestRule() )
    return ClassName[Rule[r].Rhs];
  
  return ClassName[DefaultClass];
}
    







/*************************************************************************/
/*								  	 */
/*************************************************************************/


void reset()
{
  int i;

  /*---------- ITEM ----------*/
  for( i=0 ; i<=MaxItem ; i++)
      if( Item[i] != NULL )
	{
	  cfree(Item[i]);
	  Item[i]=NULL;
	}

  if( Item != NULL )
    {
      free(Item);
      Item = NULL;
    }

  MaxItem = -1;

  if ( Rf ) fclose(Rf);
}


/*************************************************************************/
/*								  	 */
/*************************************************************************/


void init(pszPath)
char* pszPath;     
{
    Attribute a;
    int i;

    /*  Initialise  */
    FileName = pszPath;
    GetNames();

    InitialiseRules();
    GetRules("s.rules");  /* ".../DFs.rules" contains the START concepts */
    NRulesStart = NRules; 
    RuleStart = Rule;
    RuleSpaceStart = RuleSpace;
    
    InitialiseRules();
    GetRules("e.rules");  /* ".../DFe.rules" contains the START concepts */
    NRulesEnd = NRules; 
    RuleEnd = Rule;
    RuleSpaceEnd = RuleSpace;
    
    /*  Allocate value ranges  */
    RangeDesc = (struct ValRange *) calloc(MaxAtt+1, sizeof(struct ValRange));
    TestMemAlloc(RangeDesc);

    ForEach(a, 0, MaxAtt)
      if ( MaxAttVal[a] )
	{
	  RangeDesc[a].Probability = (float *) calloc(MaxAttVal[a]+1, sizeof(float));
	  TestMemAlloc(RangeDesc[a].Probability);
	}
}





char* classify( pszNewCase, cConceptType )
     char* pszNewCase;
     int cConceptType;
{
    void SetInputString();

    if ( cConceptType == 0 )
      {
	NRules = NRulesStart; 
	Rule = RuleStart;
	RuleSpace = RuleSpaceStart;
      }
    else
      {
	NRules = NRulesEnd; 
	Rule = RuleEnd;
	RuleSpace = RuleSpaceEnd;
      }

    SetInputString(pszNewCase);

    Clear();
    GetData();  

    return InterpretRuleset();
}





/************************************************************/

void OldReset()
{
  Attribute a;
  int i, j;

  for( i=0 ; i<NRulesStart ; i++ )
    {
      for( j=1 ; j<RuleStart[i].Size ; j++ )
	if( RuleStart[i].Lhs[j] != NULL )
	  {
	    free(RuleStart[i].Lhs[j]);
	    RuleStart[i].Lhs[j] = NULL;
	  }

      /*      if( RuleStart[i].Lhs != NULL )
	{
	  cfree(RuleStart[i].Lhs); 
	  RuleStart[i].Lhs = NULL;
	}*/
    }
  

  for( i=0 ; i<NRulesEnd ; i++ )
    {
      for( j=1 ; j<RuleEnd[i].Size ; j++ )
	if( RuleEnd[i].Lhs[j] != NULL )
	  {
	    free(RuleEnd[i].Lhs[j]);
	    RuleEnd[i].Lhs[j] = NULL;
	  }

      /*      if( RuleEnd[i].Lhs != NULL )
	{
	  free(RuleEnd[i].Lhs); 
	  RuleEnd[i].Lhs = NULL;
	}*/
    }

  if( RuleStart != NULL )
    free(RuleStart);

  if( RuleEnd != NULL )
    free(RuleEnd);

  RuleStart = RuleEnd = NULL;

  RuleSpaceStart = RuleSpaceEnd = 0 ;
  NRulesStart = NRulesEnd = 0; 


  InitialiseRules();



  if( TestVec != NULL )
    {
      free(TestVec);
      TestVec = NULL;
    }
  
  NTests = 0;
  TestSpace = 0;


  ForEach(a, 0, MaxAtt)
    if ( MaxAttVal[a] )
      {
	cfree(RangeDesc[a].Probability);
	RangeDesc[a].Probability = NULL;
      }

  cfree(RangeDesc);
  RangeDesc = NULL;



  /*---------- ITEM ----------*/
  for( i=0 ; i<=MaxItem ; i++)
    {
      free(Item[i]);
      Item[i]=NULL;
    }

  if( Item != NULL )
    {
      free(Item);
      Item = NULL;
    }

  MaxItem = 0;


  /*---------- ClassName ----------*/
  for( i=0 ; i<MaxClass ; i++)
    {
      free(ClassName[i]);
      ClassName[i]=NULL;
    }

  if( ClassName != NULL)
    {
      free(ClassName);
      ClassName = NULL;
    }
  
  MaxClass = -1;
  AttCeiling=100;
  ClassCeiling=100;



  /*---------- AttName ----------*/
  for( i=0 ; i<MaxAtt ; i++)
    {
      free(AttName[i]);
      AttName[i]=NULL;
    }

  if( AttName != NULL)
    {
      free(AttName);
      AttName = NULL;
    }


  /*---------- AttValName ----------*/
  for( i=0 ; i<MaxAtt ; i++)
    {
      for( j=0 ; j<MaxAttVal[MaxAtt] ; j++)
	  if( AttValName[i][j] != NULL )
	    {
	      free(AttValName[i][j]);
	      AttValName[i][j]=NULL;
	    }

      if( AttValName[i] != NULL )
	{
	  free(AttValName[i]);
	  AttValName[i]=NULL;
	}
    }

  if( AttValName != NULL)
    {
      free(AttValName);
      AttValName = NULL;
    }

  free(MaxAttVal);
  MaxAttVal = NULL;

  free(SpecialStatus);
  SpecialStatus = NULL;

  MaxAtt = 0;
}



