/* File   : kick_dir.c
   Info   : finds the best direction to kick the ball
   Author : Ali Erdem, Gal Kaminka
   Date   : June 27, 1997 and on.
*/

#include "geometry.h"

double opponent_goal_center(void)
{
  double best, tmp;
  double goal_angle, ball_angle;

  /* calculate angle from my center-of-view to opponent-goal */
  if (visible(State[NOW].Goals[OPPONENT]))
    {
      goal_angle = State[NOW].Goals[OPPONENT].SeenInfo.dir;
    }
  else
    {
      /* opponent goal is between (field_length/2,-goal_width/2) 
	 and (field_length/2,+goal_width/2).  The center of the goal is 
	 (field_length/2,0) */
#define yy State[NOW].Players[OWN][MyNumber-1].position.pos.y 
#define xx State[NOW].Players[OWN][MyNumber-1].position.pos.x
#define th State[NOW].Players[OWN][MyNumber-1].position.dir
      goal_angle = 
	NormalizeAngle(Rad2Deg(atan2(-(yy), (FIELD_LENGTH/2)-(xx))))-(th);
#undef yy
#undef xx
#undef th
}

  /* now look at ball angle */
  if (visible(State[NOW].Ball))
    {
      ball_angle = State[NOW].Ball.SeenInfo.dir;
    }
  else
    {
      ball_angle = goal_angle;  /* no use if can't see ball! 
				   (XXX may change if hear where ball is) */
    }

  /* calc difference between them, and change best if needed */

  tmp = ball_angle - goal_angle;
  if (tmp > 180.0)
    {
      /* smallest angle is "behind me", ball is to right of goal */
      if ((360-tmp) > PROBLEM_ANGLE)
	{
	  best = /* goal_angle - (360 - (ball_angle-goal_angle)) + PROBLEM_ANGLE 
		    goal_angle - 360  + (ball_angle-goal_angle)  + PROBLEM_ANGLE
		    goal_angle - 360  + ball_angle - goal_angle + PROBLEM_ANGLE
		    */
                 ball_angle + PROBLEM_ANGLE;
	  if (best > 180.0)
	    best -= 360;
	}
      else
	best = goal_angle;
    }
  else if (tmp < -180.0)
    {
      /* smallest angle is behind me, ball is to left of goal */
      if ((360+tmp) > PROBLEM_ANGLE)
	{
	  best = ball_angle - PROBLEM_ANGLE;
	}
      if (best < -180.0)
	best += 360;
    }
  else if (abs(tmp) > PROBLEM_ANGLE)
    /* not good, ball will go through me */
    {
      
#ifdef DEBUG
      printf("correction requirerd: best dir was %7.2lf\n",best);
#endif
    
      if (tmp < 0.0) /* ball is left of goal */
	{
	  /* leftisize goal */
	  /* best -= abs(tmp+120.0); */
	  best = ball_angle + PROBLEM_ANGLE;
	}
      else
	{
	  /* ball is right of goal */
	  /* best += abs(tmp-120.0);  /* rightisize goal */
	  best = ball_angle - PROBLEM_ANGLE;
	}
    }
  else 
    {
      best = goal_angle;
    }

#ifdef DEBUG
  if ((best < -180.0) || (best > 180.0))
    {
      fprintf(stderr,"best kick=%7.1lf ball at %7.1lf goal at %7.1lf\n",best,ball_angle, goal_angle);
      fprintf(stderr,"best direction to kick=%7.2lf\n",best);
    }
#endif

   return NormalizeAngle(best);
}

