/*
 * $RCSfile: libsfsr_local.h,v $ $Revision: 1.6 $ $State: Exp $
 */
/****************************************************************************
*   File: libcforsoar_local.h
*                                                                           *
*       Copyright 1993 by Loral Advanced Distributed Simulation, Inc.       *
*                                                                           *
*               Loral Advanced Distributed Simulation, Inc.                 *
*               10 Moulton Street                                           *
*               Cambridge, MA 02138                                         *
*               617-873-3660                                                *
*                                                                           *
*	This software was developed by Loral under U. S. Government contracts *
*	and may be reproduced by or for the U. S. Government pursuant to    *
*	the copyright license under the clause at DFARS 252.227-7013        *
*	(OCT 1988).                                                         *
*                                                                           *
*       Contents: Private header file for libsafsoar
*       Created: Tue Jan  5 14:20:47 PST 1993
*       Author: jesmith
*       Remarks:                                                            *
*                                                                           *
****************************************************************************/

/* Soar stuff */
#include <soar.h>

/* #include "libsafsoar.h" */

/* ---------------------------------------------------------------------- */

/* Special Constants */

#define NUMBER_OF_AGENTS 50
#define PIOVER2 (PI/2.0)
#define NUMBER_OF_RADIOS 100

/* Suns don't have FILENAME_MAX! */
#ifndef FILENAME_MAX
#define FILENAME_MAX 255
#endif


/* Structure declarations */

/* The following structure comes from Scott Huffman's SoarSim code.       */
/* This was modified by Dave Kiersey for the IFOR effort.                 */

/* The structure for a list of attribute-value pairs.  This is used to    */
/* feed wmes to the Soar agents.                                          */

typedef struct simAV_struct {
 Symbol *attr;
 Symbol *value;        /* (WHAT ABOUT MULTIPLE VALUES? - ignore for now!) */
 wme    *wme;
 bool    changed;
 int     has_wme;
 bool    public;
 struct sim_object_struct * object;
 struct simAV_struct *next;
} simAV;

/* This structure notes the head of AV lists and the object with which    */
/* the attributes are associated.                                         */

typedef struct sim_object_struct{
  simAV *  AVlist;
  Symbol * object_id;
  int      needs_internalized;
  wme    * wme_referring_to_object;
} sim_object;

/* SOAR I/O private functions */

/* AV functions */

extern sim_object * make_sim_obj(void);
extern simAV *  AV_alloc(void);
extern Symbol * getAV(sim_object obj, Symbol *attr);
extern simAV  * getAVentry(sim_object *obj, Symbol *attr);
extern Symbol * getAVstring(sim_object obj, char *attr);
extern void     make_private_AV(sim_object *obj, char *attr);
extern simAV *  setAV(sim_object *obj, Symbol *attr, Symbol *val);
extern void     setAVint(sim_object *obj, char *attr, long val);
extern void     setAVfloat(sim_object *obj, char *attr, float val);
extern void     setAVstring(sim_object *obj, char *attr, char *val);
extern sim_object * setAVobj(sim_object * obj, char * attr_name);
extern void removeAV(sim_object *obj, char *attribute);
extern void removeAV_obj(sim_object *obj);
extern simAV *  setAVmult(sim_object *obj, Symbol *attr, Symbol *val);
extern void     setAVmult_string(sim_object *obj, char *attr, char *val);
extern sim_object * setAVmult_obj(sim_object * obj, char * attr_name);

#ifdef USE_STDARGS
extern void setAVpath_int(sim_object * obj, ...);
#else
extern void setAVpath_int();
#endif /* USE_STDARGS */

#ifdef USE_STDARGS
extern void setAVpath_string(sim_object * obj, ...);
#else
extern void setAVpath_string();
#endif /* USE_STDARGS */

/* Functions managing AV I/O */

extern void     add_sim_object_to_input_link(Symbol * id_symbol, char * link_name, sim_object *obj);
extern Symbol * get_output_value (io_wme *outputs, Symbol *id, Symbol *attr);
extern void     internalize_world_object(sim_object *obj);
extern void     release_io_symbols_simobj(sim_object * obj);
extern void     release_io_symbols_simtree(sim_object * obj);

