/*
 *----------------------------------------------------------------------
 *
 * my_output_function --
 *
 *      This function gets output from Soar's working memory.  This
 *      function serves as an example of how to obtain such output from 
 *      Soar.  This example simply obtains a squared value computed
 *      by Soar and prints the value.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Prints a squared value.
 *
 *----------------------------------------------------------------------
 */
#include "robocup.h"
#include <soar.h>
#include <sys/time.h>
#include "cmd_timing.h"

/* from general_util.c */
extern int stuff_int_into_message (char *message, int index, int value);
extern int stuff_float_into_message (char *message, int index, float value);
extern int stuff_char_into_message (char *message, int index, char *name);

extern double NormalizeAngle(double); /* kick_dir.c */

extern int send_message (Socket sock, char *message);

#define min(x,y) (((x) < (y)) ? (x) : (y))


void my_output_function (soar_callback_agent the_agent, 
			 soar_callback_data data,
			 soar_call_data call_data)
{
  agent * a;
  output_call_info * output_info;
  int mode;
  int index;
  io_wme * outputs;
  static char message[BUFSIZE1];
  Bool Sent_Command = FALSE;

  Symbol *output_link_id;
  Symbol *command_name,  *command_sym, *item_sym, *next_sym,
    *command_value_turn, 
    *command_value_dash, 
    *command_value_say, 
    *item_value_say,
    *command_value_x, *command_value_y, 
    *command_value_kick_power, *command_value_kick_direction,
    *command_value_home_x, *command_value_home_y; 


  a = (agent *) the_agent;
  output_info = (output_call_info *) call_data;
  mode = output_info->mode;
  outputs = output_info->outputs;

  switch (mode) 
    {
    case ADDED_OUTPUT_COMMAND:

      
      
      /* Here we care about is new links, not changes to old ones. */
      
      /* Fetch output_link_id (e.g. O37) so we can get values     */
      /* from it.  (io_header is a Soar-defined agent variable.)  */
      /* If no output_link is found, then there's no output.      */

      output_link_id = get_output_value (outputs, a->io_header, NIL);
      if (!output_link_id) return;

      /* Fetch "command" attribute of output link.  The command   */
      /* name is the output command from the Soar agent.  This is */
      /* is used to set up the output link */
      
      command_sym = get_io_sym_constant ("command"); 
      command_name = get_output_value (outputs, output_link_id, command_sym); 
      release_io_symbol(command_sym);
      if (!command_name) return;  
      
      /* Clean up the symbols used. */
    
      break;
      
    case MODIFIED_OUTPUT_COMMAND:
      /* First get the output id */
      output_link_id = get_output_value (outputs, a->io_header, NIL);
      
      /* if command is HomePosition communication - x */
      command_sym = get_io_sym_constant ("home_x"); 
      command_value_home_x = get_output_value (outputs, output_link_id, command_sym); 
      release_io_symbol(command_sym);
      if (command_value_home_x)
	{
	  HomePosition.pos.x = command_value_home_x->ic.value;
	  /* fprintf(stderr,"Setting home position x to %2.1lf\n", HomePosition.pos.x); */
	}

      /* if command is HomePosition communication - y */
      command_sym = get_io_sym_constant ("home_y"); 
      command_value_home_y = get_output_value (outputs, output_link_id, command_sym); 
      release_io_symbol(command_sym);
      if (command_value_home_y)
	{
	  HomePosition.pos.y = command_value_home_y->ic.value;
	  /* fprintf(stderr,"Setting home position y to %2.1lf\n",HomePosition.pos.y); */
	}

      /* If command is turn */
      command_sym = get_io_sym_constant ("turn"); 
      command_value_turn = get_output_value (outputs, output_link_id, command_sym); 
      release_io_symbol(command_sym);
      if (command_value_turn)
	{
	  wait_until_action_time();
	  /* print ("\nOutput function turn: value is %2.1f\n",
	     command_value_turn->fc.value); */
	  sprintf(message,"(turn %2.1f)\n", command_value_turn->fc.value);
	  
	  /* 
	     strcpy(message, "(turn  ");
	     index = stuff_float_into_message(message, 6, command_value_turn->fc.value);
	     message[index] = ')';
	     index++;
	     message[index] = '\n';
	     
	     index++;
	     message[index] = '\0';
	   */
/*          release_io_symbol(command_value_turn); */
	  send_message(gSock, message);
	  /* update own heading */
	  State[NOW].Players[OWN][MyNumber-1].position.dir = 
	    NormalizeAngle( State[NOW].Players[OWN][MyNumber-1].position.dir
			   +command_value_turn->fc.value);

	  Sent_Command = TRUE;
        }


      /* If command is dash */
      command_sym = get_io_sym_constant ("dash"); 
      command_value_dash = get_output_value (outputs, output_link_id, command_sym); 
      release_io_symbol(command_sym);
      if (command_value_dash)
	{
	  wait_until_action_time();
	  /* print ("\nOutput function dash: value is %d\n", 
		 command_value_dash->ic.value); */
	  sprintf(message,"(dash %d)\n",min(command_value_dash->ic.value, 100));
	  /*
	    strcpy(message, "(dash  ");
	    index = stuff_int_into_message(message, 6, min(command_value_dash->ic.value, 100));
	    message[index] = ')';
	    index++;
	    message[index] = '\n';
	    index++;
	    message[index] = '\0';
	  */
	  /*          release_io_symbol(command_value_dash); */
	  send_message(gSock, message);
 	  Sent_Command = TRUE;
       }


      /* If command is say */
      command_sym = get_io_sym_constant ("say"); 
      next_sym = get_io_sym_constant ("next"); 
      item_sym = get_io_sym_constant ("item"); 

      command_value_say = get_output_value (outputs, output_link_id, command_sym); 
      release_io_symbol(command_sym);
      
      if (command_value_say)
	{
	  wait_until_action_time();
	  /* strcpy(message, "(say ISI "); */
	  (void)sprintf(message, "(say %s ",Teamname);
	  
	  index = strlen(message);

          item_value_say = get_output_value (outputs, command_value_say, item_sym);
	  /* print ("Output function say: value is %s\n",
		 item_value_say->sc.name); */
	  /* index = stuff_char_into_message(message, 9, item_value_say->sc.name); */
	  index = stuff_char_into_message(message, index, item_value_say->sc.name);

	  command_value_say = get_output_value (outputs, command_value_say, next_sym);
	  while(command_value_say)
	    {
	      message[index] = ' ';
	      index++;
	      item_value_say = get_output_value (outputs, command_value_say, item_sym);         
	      /* print ("Output function say: value is %s\n",
		     item_value_say->sc.name); */
	      index = stuff_char_into_message(message, index, item_value_say->sc.name);
	      command_value_say = get_output_value (outputs, command_value_say, next_sym);
	    }
	  message[index] = ')';
	  index++;
	  message[index] = '\n';
	  index++;
	  message[index] = '\0';
	  send_message(gSock, message);
  	  Sent_Command = TRUE;
      }


      /* If command is move */
      command_sym = get_io_sym_constant ("move-x"); 
      command_value_x = get_output_value (outputs, output_link_id, command_sym); 
      release_io_symbol(command_sym);
      command_sym = get_io_sym_constant ("move-y"); 
      command_value_y = get_output_value (outputs, output_link_id, command_sym); 
      release_io_symbol(command_sym);
      if ((command_value_x) && (command_value_y))
	{
	  wait_until_action_time();
	  /* print ("\nOutput function move: value is %d %d\n",
		 command_value_x->ic.value, command_value_y->ic.value); */
	  
	  sprintf(message,"(move %d %d)\n",command_value_x->ic.value,command_value_y->ic.value);
	  /*
	    strcpy(message, "(move ");
	    index = stuff_int_into_message(message, 6, command_value_x->ic.value);
	    message[index] = ' '; index++;
	    index = stuff_int_into_message(message, index, command_value_y->ic.value);
	    message[index] = ')';
	    index++;
	    message[index] = '\n';
	    index++;
	    message[index] = '\0';
	  */
	  /* release_io_symbol(command_value_x); */
	  /* release_io_symbol(command_value_y);  */
	  send_message(gSock, message); 
  	  Sent_Command = TRUE;
      }

      /* If command is kick */
      command_sym = get_io_sym_constant ("kick-power"); 
      command_value_kick_power = get_output_value (outputs, output_link_id, command_sym); 
      release_io_symbol(command_sym);
      command_sym = get_io_sym_constant ("kick-direction"); 
      command_value_kick_direction = get_output_value (outputs, output_link_id, command_sym); 
      release_io_symbol(command_sym);
      if ((command_value_kick_power) && (command_value_kick_direction))
	{
	  wait_until_action_time();
	  /* print ("Output function kick: value is %d %2.1f\n",
		 command_value_kick_power->ic.value, command_value_kick_direction->fc.value); */
	  sprintf(message,"(kick %d %2.1f)\n", 
		  min(command_value_kick_power->ic.value, 100),
		  command_value_kick_direction->fc.value);

	  /*
	    strcpy(message, "(kick  ");
	    index = stuff_int_into_message(message, 6, min(command_value_kick_power->ic.value, 100));
	    message[index] = ' '; index++;
	    index = stuff_float_into_message(message, index, command_value_kick_direction->fc.value);
	    message[index] = ')';
	    index++;
	    message[index] = '\n';
	    index++;
	    message[index] = '\0';
	  */
	  /* release_io_symbol(command_value_kick_power); */
	  /* release_io_symbol(command_value_kick_direction); */
	  send_message(gSock, message);
 	  Sent_Command = TRUE;
       }

      /* make sure we note the time of the action taken, so we can
	 take another one after the appropriate delay */
      if (Sent_Command)
	start_action_timer();

      /* usleep(70000); */
    break;


  case REMOVED_OUTPUT_COMMAND:
    break;
  }


  /*  fprintf(stderr,"exiting from my_output_function\n"); */

}

